#  Copyright 2020 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
from pathlib import Path
from typing import Union
from urllib.parse import urlparse


def uri_equal(uri1: str, uri2: str) -> bool:
    """Return True if two URIs are equal."""
    if uri1 == uri2:
        return True
    parsed1 = urlparse(uri1)
    parsed2 = urlparse(uri2)
    if parsed1.scheme in ["", "file"] and parsed2.scheme in ["", "file"]:
        return (
            parsed1.netloc == parsed2.netloc and parsed1.path == parsed2.path
        )
    return False


def normalize_uri(uri: Union[str, Path]) -> str:
    """Normalize URI

    Convert a file path with "file://" schema.
    Convert an relative path to absolute path

    Parameters
    ----------
    uri : str or Path

    Return
    ------
    str
        Normalized URI with schema

    """
    if isinstance(uri, Path):
        uri = str(uri.absolute())
    parsed = urlparse(uri)
    if parsed.scheme == "":
        return "file://" + os.path.abspath(uri)
    return uri



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Geometry related PySpark UDF"""

# Third Party
from pyspark.sql.functions import udf
from pyspark.sql.types import FloatType

# Rikai
from rikai.logging import logger
from rikai.spark.types.geometry import Box2dType
from rikai.types.geometry import Box2d

__all__ = ["area", "box2d", "box2d_from_center", "box2d_from_top_left"]


@udf(returnType=Box2dType())
def box2d(coords) -> Box2d:
    """Build a Box2d from ``[xmin,ymin,xmax,ymax]`` array."""
    return Box2d(coords[0], coords[1], coords[2], coords[3])


@udf(returnType=Box2dType())
def box2d_from_center(coords) -> Box2d:
    """Build a Box2d from a center-point based coordinate array:
    ``[center_x, center_y, width, height]``.

    See Also
    --------
    :py:meth:`rikai.types.geometry.Box2d.from_center`
    """
    return Box2d.from_center(coords[0], coords[1], coords[2], coords[3])


@udf(returnType=Box2dType())
def box2d_from_top_left(coords) -> Box2d:
    """Build :py:class:`Box2d` from a top-left based coordinate array:
    ``[x0, y0, width, height]``

    Example
    -------

    `Coco dataset <https://cocodataset.org/>`_ is one public dataset that use
    top-left coordinates


    >>> #! pyspark --packages ai.eto:rikai_2.12:0.0.1
    >>> import json
    >>> from rikai.spark.functions.geometry import box2d_from_top_left
    >>>
    >>> with open("coco_sample/annotations/train_sample.json") as fobj:
    ...     coco = json.load(fobj)
    >>> anno_df = (
    ...     spark
    ...     .createDataFrame(coco["annotations"][:5])
    ...     .withColumn("box2d", box2d_from_top_left("bbox"))
    ... )
    >>> anno_df.show()
    +--------------------+-----------+--------+--------------------+
    |                bbox|category_id|image_id|               box2d|
    +--------------------+-----------+--------+--------------------+
    |[505.24, 0.0, 47....|         72|  318219|Box2d(x=505.24, y...|
    |[470.68, 0.0, 45....|         72|  318219|Box2d(x=470.68, y...|
    |[442.51, 0.0, 43....|         72|  318219|Box2d(x=442.51, y...|
    |[380.74, 112.85, ...|         72|  554625|Box2d(x=380.74, y...|
    |[339.13, 32.99, 3...|         72|  554625|Box2d(x=339.13, y...|
    +--------------------+-----------+--------+--------------------+
    >>> anno_df.printSchema()
    root
    |-- bbox: array (nullable = true)
    |    |-- element: double (containsNull = true)
    |-- category_id: long (nullable = true)
    |-- image_id: long (nullable = true)
    |-- box2d: box2d (nullable = true)

    See Also
    --------
    :py:meth:`rikai.types.geometry.Box2d.from_top_left`.
    """
    return Box2d.from_top_left(coords[0], coords[1], coords[2], coords[3])


@udf(returnType=FloatType())
def area(bbox: Box2d) -> float:
    """A UDF to calculate the area of a bounding box."""
    return bbox.area



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import random
from typing import Generic, Optional, TypeVar

__all__ = ["RandomShuffler"]

Elem = TypeVar("Elem")


class RandomShuffler(Generic[Elem]):
    """Reservoir sampling-based shuffler to provide randomized access over elements.

    :py:class:`RandomShuffler` maintains an internal buffer, and uses `reservoir sampling`_
    to offer randomness with uniform distribution. The buffer ``capacity`` does
    not affect the possibility distribution.

    Parameters
    ----------
    capacity : int, optional
        The capacity of the internal random access buffer. Note that setting this value to
        ``1`` or ``0`` makes this :py:class:`RandomShuffler` to a FIFO queue. Default
        value: ``32``.
    seed : int, optional
        Random seed.

    Example
    -------

    .. code-block:: python

        def __iter__(self):
            \"\"\"Provide random access over a Stream\"\"\"
            shuffler = RandomShuffler(capacity=128)
            for elem in stream:
                shuffler.append(elem)
                # Approximately maintain the shuffler at its capacity.
                while shuffler.full():
                    yield shuffler.pop()
            while shuffler:
                yield shuffler.pop()

    Notes
    -----
    - Set ``capacity`` to ``1`` or ``0``, makes :py:class:`RandomShuffler` a FIFO queue.
    - This class is not thread-safe.

    References
    ----------
    - `Reservoir Sampling`_
    - Petastorm `Shuffling Buffer <https://github.com/uber/petastorm/blob/master/petastorm/reader_impl/shuffling_buffer.py>`_

    .. _Reservoir Sampling: https://en.wikipedia.org/wiki/Reservoir_sampling
    """  # noqa

    DEFAULT_CAPACITY = 32

    def __init__(
        self, capacity: int = DEFAULT_CAPACITY, seed: Optional[int] = None
    ):
        """Construct a :py:class:`RandomShuffler`"""
        self.capacity = capacity
        self.seed = seed
        self.buffer = []
        random.seed(self.seed)

    def __repr__(self) -> str:
        return "RandomShuffler(capacity={})".format(self.capacity)

    def __len__(self) -> int:
        """Returns the number of elements in the shuffler."""
        return len(self.buffer)

    def __bool__(self) -> bool:
        """Return True if this shuffler is not empty."""
        return len(self.buffer) > 0

    def full(self) -> bool:
        """Return True if this shuffler reaches to its capacity."""
        return len(self) >= self.capacity

    def append(self, elem: Elem):
        """Append a new element to the shuffler"""
        self.buffer.append(elem)

    def pop(self) -> Elem:
        """Pop out one random element from the shuffler.

        Raises
        ------
        IndexError
            If the internal buffer is empty.
        """
        if len(self.buffer) == 0:
            raise IndexError("Buffer is empty")
        idx = random.randrange(len(self.buffer))
        item = self.buffer[idx]
        self.buffer[idx] = self.buffer[-1]
        self.buffer.pop()
        return item



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import secrets
from typing import Any, Callable, Dict, IO, Mapping, Optional, Union

import yaml
from jsonschema import validate, ValidationError
from pyspark.sql import SparkSession

from rikai.internal.reflection import find_class
from rikai.io import open_uri
from rikai.logging import logger
from rikai.spark.sql.codegen.base import Registry
from rikai.spark.sql.exceptions import SpecError
from rikai.spark.sql.schema import parse_schema

__all__ = ["FileSystemRegistry"]

# YAML-Spec SCHEMA
SPEC_SCHEMA = {
    "type": "object",
    "properties": {
        "version": {
            "type": "number",
            "description": "Model SPEC format version",
        },
        "name": {"type": "string", "description": "Model name"},
        "schema": {"type": "string"},
        "model": {
            "type": "object",
            "description": "model description",
            "properties": {
                "uri": {"type": "string"},
                "flavor": {"type": "string"},
            },
            "required": ["uri"],
        },
        "transforms": {
            "type": "object",
            "properties": {
                "pre": {"type": "string"},
                "post": {"type": "string"},
            },
        },
    },
    "required": ["version", "schema", "model"],
}


class ModelSpec:
    """Model Spec.

    Parameters
    ----------
    spec : str, bytes, dict or file object
        String content of the serialized spec, or a dict
    options : Dict[str, Any], optional
        Additionally options. If the same option exists in spec already,
        it will be overridden.
    validate : bool, default True.
        Validate the spec during construction. Default ``True``.
    """

    def __init__(
        self,
        spec: Union[bytes, str, IO, Dict[str, Any]],
        options: Optional[Dict[str, Any]] = None,
        validate: bool = True,
    ):
        if not isinstance(spec, Mapping):
            spec = yaml.load(spec, Loader=yaml.FullLoader)
        self._spec = spec
        self._options: Dict[str, Any] = self._spec.get("options", {})
        if options:
            self._options.update(options)

        if validate:
            self.validate()

    def validate(self):
        """Validate model spec

        Raises
        ------
        SpecError
            If the spec is not well-formatted.
        """
        logger.debug("Validating spec: %s", self._spec)
        try:
            validate(instance=self._spec, schema=SPEC_SCHEMA)
        except ValidationError as e:
            raise SpecError(e.message) from e

    @property
    def version(self):
        """Returns spec version."""
        return self._spec["version"]

    @property
    def uri(self):
        """Return Model URI"""
        return self._spec["model"]["uri"]

    @property
    def flavor(self):
        """Model flavor"""
        return self._spec["model"].get("flavor", "")

    @property
    def schema(self):
        """Return the output schema of the model."""
        return parse_schema(self._spec["schema"])

    @property
    def options(self):
        """Model options"""
        return self._options

    @property
    def pre_processing(self) -> Optional[Callable]:
        """Return pre-processing transform if exists"""
        if (
            "transforms" not in self._spec
            or "pre" not in self._spec["transforms"]
        ):
            return None
        f = find_class(self._spec["transforms"]["pre"])
        return f

    @property
    def post_processing(self) -> Optional[Callable]:
        """Return post-processing transform if exists"""
        if (
            "transforms" not in self._spec
            or "post" not in self._spec["transforms"]
        ):
            return None
        f = find_class(self._spec["transforms"]["post"])
        return f


def codegen_from_yaml(
    spark: SparkSession,
    uri: str,
    name: Optional[str] = None,
    options: Optional[Dict[str, str]] = None,
) -> str:
    """Generate code from a YAML file.

    Parameters
    ----------
    spark : SparkSession
        A live spark session
    uri : str
        the model spec URI
    name : model name
        The name of the model.
    options : dict
        Optional parameters passed to the model.

    Returns
    -------
    str
        Spark UDF function name for the generated data.
    """
    with open_uri(uri) as fobj:
        spec = ModelSpec(fobj, options=options)

    if spec.version != 1.0:
        raise SpecError(
            f"Only spec version 1.0 is supported, got {spec.version}"
        )

    if spec.flavor == "pytorch":
        from rikai.spark.sql.codegen.pytorch import generate_udf

        udf = generate_udf(
            spec.uri,
            spec.schema,
            spec.options,
            pre_processing=spec.pre_processing,
            post_processing=spec.post_processing,
        )
    else:
        raise SpecError(f"Unsupported model flavor: {spec.flavor}")

    func_name = f"{name}_{secrets.token_hex(4)}"
    spark.udf.register(func_name, udf)
    logger.info(f"Created model inference pandas_udf with name {func_name}")
    return func_name


class FileSystemRegistry(Registry):
    """FileSystem-based Model Registry"""

    def __init__(self, spark: SparkSession):
        self._spark = spark
        self._jvm = spark.sparkContext._jvm

    def __repr__(self):
        return "FileSystemRegistry"

    def resolve(self, uri: str, name: str, options: Dict[str, str]):
        logger.info(f"Resolving model {name} from {uri}")

        if uri.endswith(".yml") or uri.endswith(".yaml"):
            func_name = codegen_from_yaml(self._spark, uri, name, options)
        else:
            raise ValueError(f"Model URI is not supported: {uri}")

        model = self._jvm.ai.eto.rikai.sql.model.fs.FileSystemModel(
            name, uri, func_name
        )
        # TODO: set options
        return model



#  Copyright 2020 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Parquet-based Rikai Dataset that supports automatically UDT conversions.
"""

# Standard Library
import importlib
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

# Third Party
import pyarrow.parquet as pg
from pyarrow.fs import FileSystem
from pyspark.ml.linalg import Matrix, Vector
from pyspark.sql import Row
from pyspark.sql.types import UserDefinedType

# Rikai
from rikai.io import open_input_stream
from rikai.logging import logger
from rikai.parquet.resolver import Resolver
from rikai.parquet.shuffler import RandomShuffler

__all__ = ["Dataset"]


class Dataset:
    """Rikai Dataset.

    :py:class:`Dataset` provides read access to a Rikai dataset. It supports

    - Read Rikai encoded dataset on the supported storage medias, i.e.,
      local filesystem, AWS S3 or Google GCS.
    - Automatically deserialize data into semantic user defined types (UDT).
    - Shuffle the dataset randomly through ``shuffle`` flag.
    - Distributed training by setting ``world_size`` and ``rank`` parameters.
      When enabled, parquet `row-group` level partition will be used to
      distribute data amount the workers.

    Parameters
    ----------
    query : str
        A SQL query "SELECT image, annotation FROM s3://foo/bar" or
        a dataset URI, i.e., "s3://foo/bar"
    columns : List[str], optional
        To read only given columns
    shuffle : bool, optional
        Set to True to shuffle the results.
    shuffler_capacity : int
        The size of the buffer to shuffle the examples. The size of buffer
        does not impact the distribution of possibility that an example
        is picked.
    seed : int, optional
        Random seed for shuffling process.
    world_size : int
        Total number of distributed workers
    rank : int
        The rank of this worker in all the distributed workers

    Notes
    -----
    - Typically user should not directly use this class. Instead users are
      encouraged to use framework-native readers, for example, using
      :py:class:`rikai.torch.data.DataLoader` in
      `Pytorch <https://pytorch.org/>`_
    """

    SPARK_PARQUET_ROW_METADATA = b"org.apache.spark.sql.parquet.row.metadata"

    _UDT_CACHE: Dict[str, UserDefinedType] = {}

    def __init__(  # pylint: disable=too-many-arguments
        self,
        query: Union[str, Path],
        columns: Optional[List[str]] = None,
        shuffle: bool = False,
        shuffler_capacity: int = 128,
        seed: Optional[int] = None,
        world_size: int = 1,
        rank: int = 0,
    ):
        self.uri = str(query)
        self.columns = columns
        self.shuffle = shuffle
        self.shuffler_capacity = shuffler_capacity
        self.seed = seed
        self.rank = rank
        self.world_size = world_size
        if self.world_size > 1:
            logger.info(
                "Running in distributed mode, world size=%s, rank=%s",
                world_size,
                rank,
            )

        # Provide deterministic order between distributed workers.
        self.files = sorted(Resolver.resolve(self.uri))
        if self.rank == 0:
            logger.info("Loading parquet files: %s", self.files)

        self.spark_row_metadata = Resolver.get_schema(self.uri)

    def __repr__(self) -> str:
        return "Dataset(uri={}, columns={}, shuffle={})".format(
            self.uri, self.columns if self.columns else "[*]", self.shuffle
        )

    @classmethod
    def _find_udt(cls, pyclass: str) -> UserDefinedType:
        """Find UDT class specified by the python class path."""
        if pyclass in cls._UDT_CACHE:
            return cls._UDT_CACHE[pyclass]

        class_path = pyclass.split(".")
        module_name = ".".join(class_path[:-1])
        class_name = class_path[-1]
        try:
            udt_class = getattr(
                importlib.import_module(module_name), class_name
            )
        except ImportError as err:
            raise ImportError(
                f"Could not import user defind type {pyclass}"
            ) from err
        cls._UDT_CACHE[pyclass] = udt_class()
        return cls._UDT_CACHE[pyclass]

    def _convert(
        self, raw_row: Dict[str, Any], schema: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Convert Spark UDT to native rikia or numpy types.

        Parameters
        ----------
        raw_row : Dict[str, Any]
        schema : Dict[str, Any]
            Spark schema in the JSON format
        """
        assert schema["type"] == "struct"
        converted = {}
        for field in schema["fields"]:
            name = field["name"]
            if name not in raw_row:
                # This column is not selected, skip
                continue
            field_type = field["type"]
            if isinstance(field_type, dict) and field_type["type"] == "udt":
                udt = self._find_udt(field_type["pyClass"])

                value = raw_row[name]
                if isinstance(value, dict):
                    row = Row(**value)
                else:
                    row = Row(value)
                converted_value = udt.deserialize(row)
                if isinstance(converted_value, (Vector, Matrix)):
                    # For pyspark.ml.linalg.{Vector,Matrix}, we eagly convert
                    # them into numpy ndarrays.
                    converted_value = converted_value.toArray()
                converted[name] = converted_value
            elif (
                isinstance(field_type, dict)
                and field_type["type"] == "array"
                and isinstance(field_type["elementType"], dict)
                and field_type["elementType"]["type"] in {"struct", "udt"}
            ):
                converted[name] = [
                    self._convert(elem, field_type["elementType"])
                    for elem in raw_row[name]
                ]
            else:
                converted[name] = raw_row[name]

        return converted

    def __iter__(self):
        shuffler = RandomShuffler(
            self.shuffler_capacity if self.shuffle else 1, self.seed
        )
        group_count = -1
        for file_uri in self.files:
            with open_input_stream(file_uri) as fobj:
                parquet = pg.ParquetFile(fobj)
                for group_idx in range(parquet.num_row_groups):
                    # A simple form of row-group level bucketing without
                    # memory overhead.
                    # Pros:
                    #  - It requires zero communication to initialize the
                    #    distributed policy
                    #  - It uses little memory and no startup overhead, i.e.
                    #    collecting row groups.
                    # Cons:
                    #   The drawback would be if the world size is much larger
                    #   than the average number of row groups. As a result,
                    #   many of the file open operations would be wasted.
                    group_count += 1
                    if group_count % self.world_size != self.rank:
                        continue
                    row_group = parquet.read_row_group(
                        group_idx, columns=self.columns
                    )
                    for (
                        batch
                    ) in row_group.to_batches():  # type: pyarrow.RecordBatch
                        # TODO: read batches not using pandas
                        for _, row in batch.to_pandas().iterrows():
                            shuffler.append(row)
                            # Maintain the shuffler buffer around its capacity.

                            while shuffler.full():
                                yield self._convert(
                                    shuffler.pop().to_dict(),
                                    self.spark_row_metadata,
                                )
        while shuffler:
            yield self._convert(
                shuffler.pop().to_dict(), self.spark_row_metadata
            )



#  Copyright (c) 2021 Rikai Authors
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pyspark.sql import DataFrame

from .conf import CONF_PARQUET_BLOCK_SIZE, DEFAULT_ROW_GROUP_SIZE_BYTES


def df_to_rikai(
    df: DataFrame,
    uri: str,
    parquet_row_group_size_bytes: int = DEFAULT_ROW_GROUP_SIZE_BYTES,
):
    (
        df.write.format("rikai")
        .option(CONF_PARQUET_BLOCK_SIZE, parquet_row_group_size_bytes)
        .save(uri)
    )



#  Copyright 2020 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Extensive Dataset Resolver."""

import json
import os
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Dict, Iterable, Union
from urllib.parse import urlparse

import pyarrow.parquet as pq
from pyarrow.fs import FileSelector, FileSystem

from rikai.internal.uri_utils import normalize_uri
from rikai.io import _gcsfs, open_input_stream
from rikai.logging import logger

__all__ = ["register", "Resolver", "BaseResolver"]


class BaseResolver(ABC):
    """Abstract base class for the concrete resolver"""

    @abstractmethod
    def resolve(self, uri: str) -> Iterable[str]:
        """Resolve the name of a feature dataset URI, and returns a list of
        parquet file URIs.

        """

    @abstractmethod
    def get_schema(self, uri: str):
        """Return the schema of the dataset, specified by URI."""


class DefaultResolver(BaseResolver):
    """DefaultResolver supports features on local filesystem or s3.

    Supported URIs

    - "/path/to/dataset"
    - "file://path/to/dataset"
    - "s3://path/to/dataset"
    """

    SPARK_PARQUET_ROW_METADATA = b"org.apache.spark.sql.parquet.row.metadata"

    def resolve(self, uri: str) -> Iterable[str]:
        """Resolve dataset via a filesystem URI.

        Parameters
        ----------
        uri : str
            The directory / base uri for a dataset.

        Returns
        -------
        Iterator[str]
            An iterator of parquet files.
        """
        uri = normalize_uri(uri)
        parsed = urlparse(uri)
        scheme = parsed.scheme

        paths = None
        if scheme == "gs":
            fs = _gcsfs()
            glob_uri = os.path.join(uri, "*.parquet")
            logger.debug("Scan GCS directory: %s", glob_uri)
            paths = fs.glob(glob_uri)
        else:
            logger.debug("Scan pyarrow supported directory: %s", uri)
            fs, base_dir = FileSystem.from_uri(uri)
            # base_dir = parsed.netloc + parsed.path
            selector = FileSelector(
                base_dir, allow_not_found=True, recursive=True
            )
            scheme = parsed.scheme if parsed.scheme else "file"
            paths = (
                finfo.path
                for finfo in fs.get_file_info(selector)
                if finfo.path.endswith(".parquet")
            )
        return (scheme + "://" + path for path in paths)

    def get_schema(self, uri: str):
        """Get the schema of the dataset.

        Parameters
        ----------
        uri : str
            The directory URI for the dataset.

        Returns
        -------
        Dict
            Json formatted schema of the dataset.
        """

        first_parquet = next(self.resolve(uri))
        logger.debug("Resolve dataset schema from %s", first_parquet)
        metadata_file = open_input_stream(str(first_parquet))
        metadata = pq.read_metadata(metadata_file)

        kv_metadata = metadata.metadata
        try:
            return json.loads(kv_metadata[self.SPARK_PARQUET_ROW_METADATA])
        except KeyError as exp:
            raise ValueError(
                f"Parquet dataset {uri} is not created via Spark"
            ) from exp


class Resolver:
    """Extensible Dataset Resolver"""

    _DEFAULT_SCHEME = "_DEFAULT"
    # Mapping from scheme to a daaset resolver
    _RESOLVERS: Dict[str, BaseResolver] = {_DEFAULT_SCHEME: DefaultResolver()}

    @classmethod
    def reset(cls):
        """Reset Resolver for testing purpose."""
        cls._RESOLVERS = {cls._DEFAULT_SCHEME: DefaultResolver()}

    @classmethod
    def register(cls, scheme: str, resolver: BaseResolver):
        """
        Register a customize dataset resolver with given scheme, providing
        integration with feature store registeration.

        Parameters
        ----------
        scheme : str
            Feature Dataset URI scheme
        resolver : BaseResolver
            Parquet file resolver

        Raises
        ------
        KeyError
            If the same scheme name has already been registered
        """
        if scheme in cls._RESOLVERS:
            raise KeyError(f"scheme f{scheme} has already been registered")
        cls._RESOLVERS[scheme] = resolver()

    @classmethod
    def resolve(cls, uri: Union[str, Path]) -> Iterable[str]:
        """Resolve the dataset URI, and returns a list of parquet files."""
        uri = str(uri)
        parsed = urlparse(uri)
        if parsed.scheme in cls._RESOLVERS:
            logger.debug("Use extended resolver for scheme: %s", parsed.scheme)
            return cls._RESOLVERS[parsed.scheme].resolve(uri)
        return cls._RESOLVERS[cls._DEFAULT_SCHEME].resolve(uri)

    @classmethod
    def get_schema(cls, uri: str):
        """Get the schema of the dataset

        Parameters
        ----------
        uri : str
            URI of the dataset
        """
        parsed = urlparse(uri)
        if parsed.scheme in cls._RESOLVERS:
            logger.debug("Use extended resolver for scheme: %s", parsed.scheme)
            return cls._RESOLVERS[parsed.scheme].get_schema(uri)
        return cls._RESOLVERS[cls._DEFAULT_SCHEME].get_schema(uri)


def register(scheme: str):
    """
    A decorator that registers a customize dataset resolver with given scheme,
    providing integration with featurestore registeration.

    Parameters
    ----------
    scheme : str
        Feature Dataset URI scheme

    Raises
    ------
    ValueError
        If the same scheme name has already been registered

    Examples
    --------

    .. code-block:: python

        # Extend Rikai with a smart feature store
        @register("smart")
        class SmartResolver(rikai.parquet.resolver.BaseResolver):
            def resolve(self, uri: str) -> Iterable[str]:
                return smart_client.get_files(uri)

        dataset = rikai.parquet.Dataset("smart://featureA/version/1")
    """

    def wrap(resolver: BaseResolver):
        Resolver.register(scheme, resolver)
        return resolver

    return wrap



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod
from typing import Dict


class Registry(ABC):
    """Base class of a Model Registry"""

    @abstractmethod
    def resolve(self, uri: str, name: str, options: Dict[str, str]):
        """Resolve a model from a model URI.

        Parameters
        ----------
        uri : str
            Model URI
        name : str, optional
            Optional model name. Can be empty or None. If provided, it
            overrides the model name got from the model URI.
        options: dict
            Additional options passed to the model.
        """



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pathlib import Path
from typing import Callable, Dict, Iterator, Optional, Union

import pandas as pd
import torch
from pyspark.sql.functions import pandas_udf
from pyspark.sql.types import DataType
from torch.utils.data import DataLoader

from rikai.io import open_uri
from rikai.torch.pandas import PandasDataset


def generate_udf(
    model_uri: Union[str, Path],
    schema: DataType,
    options: Optional[Dict[str, str]] = None,
    pre_processing: Optional[Callable] = None,
    post_processing: Optional[Callable] = None,
):
    """Construct a UDF to run pytorch model.

    Parameters
    ----------
    model_uri : str
        The URI pointed to a model.
    schema : DataType
        Return type of the model inference function.
    options : Dict[str, str], optional
        Runtime options

    Returns
    -------
    A Spark Pandas UDF.
    """
    options = {} if options is None else options
    use_gpu = options.get("device", "cpu") == "gpu"
    num_workers = int(options.get("num_workers", 4))
    batch_size = int(options.get("batch_size", 4))

    def torch_inference_udf(
        iter: Iterator[pd.DataFrame],
    ) -> Iterator[pd.DataFrame]:

        with open_uri(model_uri) as fobj:
            model = torch.load(fobj)
        device = torch.device("cuda" if use_gpu else "cpu")

        model.to(device)
        model.eval()

        with torch.no_grad():
            for series in iter:
                dataset = PandasDataset(series, transform=pre_processing)
                results = []
                for batch in DataLoader(
                    dataset,
                    batch_size=batch_size,
                    num_workers=num_workers,
                ):
                    predictions = model(batch)
                    if post_processing:
                        predictions = post_processing(predictions)
                    results.extend(predictions)
                yield pd.DataFrame(results)

    return pandas_udf(torch_inference_udf, returnType=schema)



#  Copyright 2020 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# Standard
import functools
import shutil
from io import BytesIO
from os.path import basename, join
from pathlib import Path
from typing import BinaryIO, IO, Union
from urllib.parse import ParseResult, urlparse

# Third Party
import requests
from pyarrow import fs

# Rikai
from rikai.logging import logger

__all__ = ["copy", "open_uri"]


def _normalize_uri(uri: str) -> str:
    parsed = urlparse(uri)
    scheme = parsed.scheme
    if not scheme:
        scheme = "file"
    elif scheme in ["s3a", "s3n"]:
        scheme = "s3"
    return ParseResult(
        scheme=scheme,
        netloc=parsed.netloc,
        path=parsed.path,
        query=parsed.query,
        fragment=parsed.fragment,
        params=parsed.params,
    ).geturl()


@functools.lru_cache(maxsize=1)
def _gcsfs(project="", token=None, block_size=None):
    try:
        import gcsfs
    except ImportError as e:
        raise ImportError(
            "Please make sure gcsfs is installed via `pip install rikai[gcp]`"
        ) from e
    return gcsfs.GCSFileSystem(
        project=project, token=token, block_size=block_size
    )


def open_input_stream(uri: str) -> BinaryIO:
    """Open a URI and returns the content as a File Object."""
    parsed = urlparse(uri)
    if parsed.scheme == "gs":
        return _gcsfs().open(uri)
    else:
        filesystem, path = fs.FileSystem.from_uri(uri)
        return filesystem.open_input_file(path)


def _open_output_stream(uri: str) -> BinaryIO:
    parsed = urlparse(uri)
    if parsed.scheme == "gs":
        return _gcsfs().open(uri, mode="wb")
    else:
        filesystem, path = fs.FileSystem.from_uri(uri)
        return filesystem.open_output_stream(path)


def copy(source: str, dest: str) -> str:
    """Copy a file from source to destination, and return the URI of
    the copied file.

    Parameters
    ----------
    source : str
        The source URI to copy from
    dest : str
        The destination uri or the destination directory. If ``dest`` is
        a URI ends with a "/", it represents a directory.

    Return
    ------
    str
        Return the URI of destination.
    """
    source = _normalize_uri(source)
    dest = _normalize_uri(dest)
    parsed_source = urlparse(source)
    if dest and dest.endswith("/"):
        dest = join(dest, basename(parsed_source.path))
    parsed_dest = urlparse(dest)
    logger.debug("Copying %s to %s", source, dest)

    if parsed_dest.scheme == parsed_source.scheme:
        # Direct copy with the same file system
        scheme = parsed_dest.scheme
        if scheme == "s3":
            s3fs, source_path = fs.FileSystem.from_uri(source)
            _, dest_path = fs.FileSystem.from_uri(dest)
            s3fs.copy(source_path, dest_path)
            return dest
        elif scheme == "gs":
            _gcsfs().copy(source, dest)
            return dest

    with _open_output_stream(dest) as out_stream, open_input_stream(
        source
    ) as in_stream:
        shutil.copyfileobj(in_stream, out_stream)
    return dest


def open_uri(uri: Union[str, Path], mode="rb") -> IO:
    """Open URI for read.

    It supports the following URI pattens:

    - File System: ``/path/to/file`` or ``file:///path/to/file``
    - AWS S3: ``s3://``
    - Google Cloud Storage: ``gs://``
    - Http(s): ``http://`` or ``https://``

    Parameters
    ----------
    uri : str or :py:class:`~pathlib.Path`
        URI of the object

    Return
    ------
    File
        A file-like object for sequential read.
    """
    if isinstance(uri, Path):
        return uri.open()
    parsed_uri = urlparse(uri)
    if not parsed_uri.scheme:
        # This is a local file
        return open(uri, mode=mode)
    elif parsed_uri.scheme in ("http", "https"):
        resp = requests.get(uri)
        return BytesIO(resp.content)
    elif parsed_uri.scheme == "gs":
        return _gcsfs().open(uri, mode=mode)
    else:
        filesystem, path = fs.FileSystem.from_uri(uri)
        return filesystem.open_input_file(path)



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Convert Coco dataset into Rikai format.

https://cocodataset.org/#home
"""

from __future__ import annotations

import json
import os
from itertools import islice
from typing import Optional

from pyspark.sql import DataFrame, SparkSession
from pyspark.sql.functions import col, lit
from pyspark.sql.types import (
    ArrayType,
    FloatType,
    IntegerType,
    LongType,
    StringType,
    StructField,
    StructType,
)

try:
    from pycocotools.coco import COCO
except ImportError as exc:
    raise ImportError("Please install pycocotools") from exc

from rikai.spark.functions import image_copy
from rikai.spark.types import Box2dType, ImageType
from rikai.types import Box2d, Image

__all__ = ["convert"]


def load_categories(annotation_file):
    """Load categories to be global categories."""
    with open(annotation_file) as fobj:
        ann = json.load(fobj)
        return {c["id"]: c for c in ann["categories"]}


def convert(
    spark: SparkSession,
    dataset_root: str,
    limit: int = 0,
    asset_dir: Optional[str] = None,
) -> DataFrame:
    """Convert a Coco Dataset into Rikai dataset.

    This function expects the COCO datasets are stored in directory with the
    following structure:

    - dataset
        - annotations
          - captions_train2017.json
          - instances_train2017.json
          - ...
        - train2017
        - val2017
        - test2017

    Parameters
    ----------
    spark : SparkSession
        A live spark session
    dataset_root : str
        The directory of dataset
    limit : int, optional
        The number of images of each split to be converted.
    asset_dir : str, optional
        The asset directory to store images, can be a s3 directory.

    Return
    ------
    DataFrame
        Returns a Spark DataFrame
    """
    train_json = os.path.join(
        dataset_root, "annotations", "instances_train2017.json"
    )
    val_json = os.path.join(
        dataset_root, "annotations", "instances_val2017.json"
    )

    categories = load_categories(train_json)

    examples = []
    for split, anno_file in zip(["train", "val"], [train_json, val_json]):
        coco = COCO(annotation_file=anno_file)
        # Coco has native dependencies, so we do not distributed them
        # to the workers.
        image_ids = coco.imgs
        if limit > 0:
            image_ids = islice(image_ids, limit)
        for image_id in image_ids:
            ann_id = coco.getAnnIds(imgIds=image_id)
            annotations = coco.loadAnns(ann_id)
            annos = []
            for ann in annotations:
                bbox = Box2d(*ann["bbox"])
                annos.append(
                    {
                        "category_id": ann["category_id"],
                        "category_text": categories[ann["category_id"]][
                            "name"
                        ],
                        "bbox": bbox,
                        "area": float(ann["area"]),
                    }
                )
            image_payload = coco.loadImgs(ids=image_id)[0]
            example = {
                "image_id": image_id,
                "annotations": annos,
                "image": Image(
                    os.path.abspath(
                        os.path.join(
                            os.curdir,
                            "dataset",
                            "{}2017".format(split),
                            image_payload["file_name"],
                        )
                    )
                ),
                "split": split,
            }
            examples.append(example)

    schema = StructType(
        [
            StructField("image_id", LongType(), False),
            StructField(
                "annotations",
                ArrayType(
                    StructType(
                        [
                            StructField("category_id", IntegerType()),
                            StructField("category_text", StringType()),
                            StructField("area", FloatType()),
                            StructField("bbox", Box2dType()),
                        ]
                    )
                ),
                False,
            ),
            StructField("image", ImageType(), False),
            StructField("split", StringType(), False),
        ]
    )
    df = spark.createDataFrame(examples, schema=schema)
    if asset_dir:
        asset_dir = asset_dir if asset_dir.endswith("/") else asset_dir + "/"
        print("ASSET DIR: ", asset_dir)
        df = df.withColumn("image", image_copy(col("image"), lit(asset_dir)))
    return df



#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Vision Related User-defined Types:

- :py:class:`Image`
"""

from __future__ import annotations

from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Union
from urllib.parse import urlparse

# Third-party libraries
import numpy as np
from PIL import Image as PILImage

# Rikai
from rikai.internal.uri_utils import normalize_uri
from rikai.io import copy
from rikai.mixin import Asset, Displayable, ToNumpy, ToPIL
from rikai.spark.types import ImageType

__all__ = ["Image"]


class Image(ToNumpy, ToPIL, Asset, Displayable):
    """An external Image Asset.

    It contains a reference URI to an image stored on the remote system.

    Parameters
    ----------
    uri : str
        The URI pointed to an Image stored on external storage.
    """

    __UDT__ = ImageType()

    def __init__(self, uri: Union[str, Path]):
        super().__init__(uri)
        self._cached_data = None

    @classmethod
    def from_array(
        cls,
        array: np.ndarray,
        uri: Union[str, Path],
        mode: str = None,
    ) -> Image:
        """Create an image in memory from numpy array.

        Parameters
        ----------
        array : np.ndarray
            Array data
        uri : str or Path
            The external URI to store the data.
        mode : str, optional
            The mode which PIL used to create image. See supported
            `modes on PIL document <https://pillow.readthedocs.io/en/stable/handbook/concepts.html#concept-modes>`_.

        See Also
        --------
        :py:class:`PIL.Image.fromarray`
        :py:func:`~rikai.spark.functions.vision.numpy_to_image`

        """  # noqa: E501

        assert array is not None
        img = PILImage.fromarray(array, mode=mode)
        return cls.from_pil(img, uri)

    @staticmethod
    def from_pil(img: PILImage, uri: Union[str, Path]) -> Image:
        """Create an image in memory from a :py:class:`PIL.Image`.

        Parameters
        ----------
        img : :py:class:`PIL.Image`
            An PIL Image instance
        uri : str or Path
            The URI to store the image externally.
        """
        parsed = urlparse(normalize_uri(uri))
        if parsed.scheme == "file":
            img.save(uri)
        else:
            with NamedTemporaryFile() as fobj:
                img.save(fobj)
                fobj.flush()
                copy(fobj.name, uri)
        return Image(uri)

    def display(self, **kwargs):
        """
        Custom visualizer for this image in jupyter notebook

        Parameters
        ----------
        kwargs: dict
            Optional display arguments

        Returns
        -------
        img: IPython.display.Image
        """
        from IPython.display import Image

        with self.open() as fobj:
            return Image(fobj.read(), **kwargs)

    def __repr__(self) -> str:
        return f"Image(uri={self.uri})"

    def _repr_html_(self):
        """Default visualizer for remote ref (or local ref under cwd)"""
        return self.display()._repr_html_()

    def _repr_mimebundle_(self, include=None, exclude=None):
        """default visualizer for embedded mime bundle"""
        return self.display()._repr_mimebundle_(
            include=include, exclude=exclude
        )

    def _repr_jpeg_(self):
        """default visualizer for embedded jpeg"""
        return self.display()._repr_jpeg_()

    def _repr_png_(self):
        """default visualizer for embedded png"""
        return self.display()._repr_png_()

    def __eq__(self, other) -> bool:
        return isinstance(other, Image) and super().__eq__(other)

    def to_pil(self) -> PILImage:
        """Return an PIL image.

        Note
        ----
        The caller should close the image.
        https://pillow.readthedocs.io/en/stable/reference/open_files.html#image-lifecycle
        """
        return PILImage.open(self.open())

    def to_numpy(self) -> np.ndarray:
        """Convert this image into an :py:class:`numpy.ndarray`."""
        if self._cached_data is None:
            with self.to_pil() as pil_img:
                self._cached_data = np.asarray(pil_img)
        assert self._cached_data is not None
        return self._cached_data



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pyspark.sql.types import (
    IntegerType,
    StringType,
    StructField,
    StructType,
    UserDefinedType,
)

__all__ = ["YouTubeVideoType", "VideoStreamType"]


class VideoStreamType(UserDefinedType):
    """VideoStreamType defines the Spark UserDefineType for
    a given video stream
    """

    def __init__(self):
        super().__init__()

    def __repr__(self) -> str:
        return "VideoType"

    @classmethod
    def sqlType(cls) -> StructType:
        return StructType(
            fields=[StructField("uri", StringType(), nullable=False)]
        )

    @classmethod
    def module(cls) -> str:
        return "rikai.spark.types"

    @classmethod
    def scalaUDT(cls) -> str:
        return "org.apache.spark.sql.rikai.VideoStreamType"

    def serialize(self, obj: "VideoStream"):
        """Serialize a VideoStream to a Spark Row"""
        return (obj.uri,)

    def deserialize(self, datum) -> "VideoStream":
        from rikai.types import (
            VideoStream,
        )  # pylint: disable=import-outside-toplevel

        return VideoStream(datum[0])

    def simpleString(self) -> str:
        return "VideoStreamType"


class YouTubeVideoType(UserDefinedType):
    """YouTubeVideoType defines the Spark UserDefineType for
    a piece of YouTube video content (i.e., corresponds to a given
    youtube id but can have multiple streams)
    """

    def __init__(self):
        super().__init__()

    def __repr__(self) -> str:
        return "YouTubeVideoType"

    @classmethod
    def sqlType(cls) -> StructType:
        return StructType(
            fields=[StructField("vid", StringType(), nullable=False)]
        )

    @classmethod
    def module(cls) -> str:
        return "rikai.spark.types"

    @classmethod
    def scalaUDT(cls) -> str:
        return "org.apache.spark.sql.rikai.YouTubeVideoType"

    def serialize(self, obj: "YouTubeVideo"):
        """Serialize a YouTubeVideo"""
        return (obj.vid,)

    def deserialize(self, datum) -> "YouTubeVideo":
        from rikai.types import (
            YouTubeVideo,
        )  # pylint: disable=import-outside-toplevel

        return YouTubeVideo(datum[0])

    def simpleString(self) -> str:
        return "YouTubeVideoType"


class SegmentType(UserDefinedType):
    """Defines the Spark UDT for a video segment
    A video segment is defined by the starting frame number (start_fno)
    and ending frame number (end_fno). `fno` terminology is chosen to be
    consistent with the opencv library.
    """

    def __init__(self):
        super().__init__()

    def __repr__(self) -> str:
        return "SegmentType"

    @classmethod
    def sqlType(cls) -> StructType:
        return StructType(
            fields=[
                StructField("start_fno", IntegerType(), nullable=False),
                StructField("end_fno", IntegerType(), nullable=False),
            ]
        )

    @classmethod
    def module(cls) -> str:
        return "rikai.spark.types"

    @classmethod
    def scalaUDT(cls) -> str:
        return "org.apache.spark.sql.rikai.SegmentType"

    def serialize(self, obj: "Segment"):
        return (obj.start_fno, obj.end_fno)

    def deserialize(self, datum) -> "Segment":
        from rikai.types import (
            Segment,
        )  # pylint: disable=import-outside-toplevel

        return Segment(datum[0], datum[1])

    def simpleString(self) -> str:
        return "SegmentType"



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Dict

from pyspark.sql import SparkSession


class TestModel:
    def __init__(self, name: str, uri: str, options: Dict[str, str]):
        self.name = name
        self.uri = uri
        self.options = options

    def codegen(self, spark: SparkSession, temporary: bool):
        """Codegen for :py:class:`TestModel`

        Parameters
        ----------
        spark : SparkSession
            SparkSession

        temporary : bool
            Whether this model is generate temporary functions.
        """
        pass



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, Mapping

import numpy as np
import pandas as pd

from rikai.mixin import ToNumpy, ToPIL

__all__ = ["RikaiToTensor"]


def convert_tensor(row, use_pil: bool = False):
    """Convert a parquet row into rikai semantic objects."""
    if not isinstance(row, (Mapping, pd.Series)):
        # Primitive values
        return row
    tensors = {}
    for key, value in row.items():
        if isinstance(value, dict):
            tensors[key] = convert_tensor(value)
        elif isinstance(value, (list, tuple)):
            tensors[key] = np.array([convert_tensor(elem) for elem in value])
        elif use_pil and isinstance(value, ToPIL):
            tensors[key] = value.to_pil()
        elif isinstance(value, ToNumpy):
            tensors[key] = value.to_numpy()
        else:
            tensors[key] = value
    return tensors


class RikaiToTensor:
    """Convert a Row in the ``Rikai`` parquet dataset into Pytorch Tensors

    Warnings
    --------
    Internal use only
    """

    def __init__(self, use_pil: bool = False):
        self.use_pil = use_pil

    def __repr__(self) -> str:
        return "ToTensor"

    def __call__(self, record) -> Any:
        return convert_tensor(record, use_pil=self.use_pil)



#  Copyright (c) 2021 Rikai Authors
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# for parquet.block.size when caching during df -> Dataset conversion
CONF_PARQUET_BLOCK_SIZE = "parquet.block.size"
DEFAULT_ROW_GROUP_SIZE_BYTES = 32 * 1024 * 1024

# config name for the cache root during df -> Dataset conversion
CONF_SPARK_RIKAI_CACHEURI = "spark.rikai.cacheUri"



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


from pyspark.sql.types import (
    StringType,
    StructField,
    StructType,
    UserDefinedType,
)

__all__ = ["ImageType"]


class ImageType(UserDefinedType):
    """ImageType defines the Spark UserDefineType for Image type"""

    def __init__(self):
        super().__init__()
        self.codec = "png"

    def __repr__(self) -> str:
        return f"ImageType(codec={self.codec})"

    @classmethod
    def sqlType(cls) -> StructType:
        return StructType(
            fields=[StructField("uri", StringType(), nullable=False)]
        )

    @classmethod
    def module(cls) -> str:
        return "rikai.spark.types.vision"

    @classmethod
    def scalaUDT(cls) -> str:
        return "org.apache.spark.sql.rikai.ImageType"

    def serialize(self, obj: "Image"):
        """Serialize an Image to a Spark Row?"""
        return (obj.uri,)

    def deserialize(self, datum) -> "Image":
        from rikai.types.vision import Image

        return Image(datum[0])

    def simpleString(self) -> str:
        return "ImageType"



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""`Torchvision`_ compatible Dataset


.. _torchvision: https://pytorch.org/vision/stable/index.html
"""

from pathlib import Path
from typing import Any, Callable, List, Optional, Tuple, Union

import PIL

import rikai.torch.data
from rikai.torch.transforms import RikaiToTensor

__all__ = ["Dataset"]


class Dataset(rikai.torch.data.Dataset):
    """A Rikai Dataset compatible with `torchvision`_.

    Parameters
    ----------
    uri_or_df : str, Path, or pyspark.sql.DataFrame
        URI of the dataset or the dataset as a pyspark DataFrame
    image_column : str
        The column name for the image data.
    target_column : str or list[str]
        The column(s) of the target / label.
    transform : Callable, optional
        A function/transform that takes in an :py:class:`PIL.Image.Image` and
        returns a transformed version. E.g,
        :py:class:`torchvision.transforms.ToTensor`
    target_transform : Callable, optional
        A function/transform that takes in the target and transforms it.

    Yields
    ------
    (image, target)

    See Also
    --------
    `Torchvision Dataset <https://pytorch.org/vision/stable/datasets.html#>`_

    Examples
    --------
    >>> from torchvision import transforms
    >>> from rikai.torch.vision import Dataset
    >>> transform = transforms.Compose(
    ...     transforms=[
    ...         transforms.Resize(128),
    ...         transforms.ToTensor(),
    ...         transforms.Normalize(
    ...             (0.485, 0.456, 0.406),
    ...             (0.229, 0.224, 0.225)
    ...         ),
    ...     ])
    >>> dataset = Dataset("out", "image", ["label"], transform=transform)
    >>> next(iter(dataset))
    ... tensor([[[-1.8610, -0.8678, -0.4226,  ..., -1.7583,  0.0569, -0.6794],
         [-1.5870, -1.8782, -1.7069,  ..., -1.1075, -1.1760, -1.8782],
         [-2.1179, -0.5253, -1.7925,  ..., -0.3712, -1.4843, -1.2959],
         ...,
         [-1.1073, -0.3927, -0.8110,  ..., -0.9853,  0.1128, -1.0027]]]) dog

    """

    def __init__(
        self,
        uri_or_df: Union[str, Path, "pyspark.sql.DataFrame"],
        image_column: str,
        target_column: Union[str, List[str]],
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
    ):
        self.image_column = image_column
        self.target_columns = (
            [target_column]
            if isinstance(target_column, str)
            else target_column
        )
        super().__init__(
            uri_or_df,
            [self.image_column] + self.target_columns,
            transform=RikaiToTensor(use_pil=True),
        )

        self.transform = transform if transform else lambda x: x
        self.target_transform = (
            target_transform if target_transform else lambda x: x
        )

    def __repr__(self) -> str:
        return f"RikaiDataset({self.uri_or_df})"

    def __iter__(self) -> Tuple[PIL.Image.Image, Any]:
        for row in super().__iter__():
            image = row[self.image_column]
            target = tuple([row[col] for col in self.target_columns])
            if len(target) == 1:
                target = target[0]
            yield self.transform(image), self.target_transform(target)



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Geometry types
"""

from __future__ import annotations

from numbers import Real
from typing import List, Sequence, Tuple, Union

import numpy as np

from rikai.mixin import ToNumpy
from rikai.spark.types.geometry import Box2dType, Box3dType, PointType

__all__ = ["Point", "Box3d", "Box2d"]


class Point(ToNumpy):
    """Point in a 3-D space, specified by ``(x, y, z)`` coordinates.

    Attributes
    ----------
    x : float
        The X coordinate.
    y : float
        The Y coordinate.
    z : float
        The Z coordinate.
    """

    __UDT__ = PointType()

    def __init__(self, x: float, y: float, z: float):
        # pylint: disable=invalid-name
        self.x = float(x)
        self.y = float(y)
        self.z = float(z)

    def __repr__(self) -> str:
        return f"Point({self.x}, {self.y}, {self.z})"

    def __eq__(self, o: object) -> bool:
        return (
            isinstance(o, Point)
            and self.x == o.x
            and self.y == o.y
            and self.z == o.z
        )

    def to_numpy(self) -> np.ndarray:
        return np.array([self.x, self.y, self.z])


class Box2d(ToNumpy, Sequence):
    """2-D Bounding Box, defined by ``(xmin, ymin, xmax, ymax)``

    Attributes
    ----------
    xmin : float
        X-coordinate of the top-left point of the box.
    ymin : float
        Y-coordinate of the top-left point of the box.
    xmax : float
        X-coordinate of the bottm-right point of the box.
    ymax : float
        Y-coordinate of the bottm-right point of the box.

    Example
    -------

    >>> box = Box2d(1, 2, 3, 4)
    >>> box / 2
    Box2d(xmin=0.5, ymin=1.0, xmax=1.5, ymax=2.0)
    >>> box * (3.5, 5)
    Box2d(xmin=3.5, ymin=10.0, xmax=10.5, ymax=20.0)
    >>> # Box2d can be used directly with PIL.ImageDraw
    >>> draw = PIL.ImageDraw.Draw(img)
    >>> draw.rectangle(box, fill="green", width=2)
    """

    __UDT__ = Box2dType()

    def __init__(self, xmin: float, ymin: float, xmax: float, ymax: float):
        assert (
            0 <= xmin <= xmax
        ), f"xmin({xmin}) and xmax({xmax}) must satisfy 0 <= xmin <= xmax"
        assert (
            0 <= ymin <= ymax
        ), f"ymin({ymin}) and ymax({ymax}) must satisfy 0 <= ymin <= ymax"
        self.xmin = float(xmin)
        self.ymin = float(ymin)
        self.xmax = float(xmax)
        self.ymax = float(ymax)

    @classmethod
    def from_center(
        cls, center_x: float, center_y: float, width: float, height: float
    ) -> Box2d:
        """Factory method to construct a :py:class:`Box2d` from
        the center point coordinates: ``{center_x, center_y, width, height}``.


        Parameters
        ----------
        center_x : float
            X-coordinate of the center point of the box.
        center_y : float
            Y-coordinate of the center point of the box.
        width : float
            The width of the box.
        height : float
            The height of the box.

        Return
        ------
        Box2d
        """
        assert (
            width >= 0 and height >= 0
        ), f"Box2d width({width}) and height({height}) must be non-negative."
        return Box2d(
            center_x - width / 2,
            center_y - height / 2,
            center_x + width / 2,
            center_y + height / 2,
        )

    @classmethod
    def from_top_left(
        cls, xmin: float, ymin: float, width: float, height: float
    ) -> Box2d:
        """Construct a :py:class:`Box2d` from
        the top-left based coordinates: ``{x0, y0, width, height}``.

        Top-left corner of an image / bbox is `(0, 0)`.

        Several public datasets, including `Coco Dataset`_, use this
        coordinations.

        Parameters
        ----------
        xmin : float
            X-coordinate of the top-left point of the box.
        ymin : float
            Y-coordinate of the top-left point of the box.
        width : float
            The width of the box.
        height : float
            The height of the box.


        References
        ----------
        - `Coco Dataset`_

        .. _Coco Dataset: https://cocodataset.org/
        """
        assert (
            width >= 0 and height >= 0
        ), f"Box2d width({width}) and height({height}) must be non-negative."
        return Box2d(xmin, ymin, xmin + width, ymin + height)

    def __repr__(self) -> str:
        return (
            f"Box2d(xmin={self.xmin}, ymin={self.ymin}, xmax={self.xmax}"
            + f", ymax={self.ymax})"
        )

    def __eq__(self, o: Box2d) -> bool:
        return isinstance(o, Box2d) and np.array_equal(
            self.to_numpy(), o.to_numpy()
        )

    def __len__(self) -> int:
        return 4

    def __getitem__(self, key: int) -> float:
        return [self.xmin, self.ymin, self.xmax, self.ymax][key]

    @staticmethod
    def _verified_scale(
        scale: Union[Real, Tuple[float, float]]
    ) -> Tuple[float, float]:
        if isinstance(scale, Real):
            assert scale > 0, f"scale must be positive, got {scale}"
            return scale, scale
        assert (
            type(scale) == tuple and len(scale) == 2
        ), f"scale must be either a number or a 2-element tuple, got {scale}"
        assert (
            scale[0] > 0 and scale[1] > 0
        ), f"scale must be positive, got {scale}"
        return scale

    def __truediv__(self, scale: Union[int, float, Tuple]) -> Box2d:
        """Scale down :py:class:`Box2d`

        Parameters
        ----------
        scale : number or a 2-element tuple/list
            Scale the Box2d by this amount.

        Return
        ------
        Box2d
            The scaled-down Box2d.

        """
        x_scale, y_scale = self._verified_scale(scale)
        return Box2d(
            xmin=self.xmin / x_scale,
            ymin=self.ymin / y_scale,
            xmax=self.xmax / x_scale,
            ymax=self.ymax / y_scale,
        )

    def __mul__(self, scale: Union[int, float, Tuple]) -> Box2d:
        """Scale up :py:class:`Box2d`

        Parameters
        ----------
        scale : number or a 2-element tuple/list
            Scale the Box2d by this amount.

        Return
        ------
        Box2d
            The scaled-up Box2d.
        """
        x_scale, y_scale = self._verified_scale(scale)
        return self / (1.0 / x_scale, 1.0 / y_scale)

    def to_numpy(self) -> np.ndarray:
        """Convert a :py:class:`Box2d` to numpy ndarray:
        ``array([xmin, ymin, xmax, ymax])``

        """
        return np.array([self.xmin, self.ymin, self.xmax, self.ymax])

    @property
    def width(self) -> float:
        return self.xmax - self.xmin

    @property
    def height(self) -> float:
        return self.ymax - self.ymin

    @property
    def area(self) -> float:
        """Area of the bounding box"""
        return self.width * self.height

    def iou(self, other: "Box2d") -> float:
        """Compute intersection over union(IOU)."""
        assert isinstance(
            other, Box2d
        ), f"Can only compute iou between Box2d, got {type(other)}"
        # Find intersection
        xmin = max(self.xmin, other.xmin)
        ymin = max(self.ymin, other.ymin)
        xmax = min(self.xmax, other.xmax)
        ymax = min(self.ymax, other.ymax)
        inter_area = max(0, xmax - xmin) * max(0, ymax - ymin)

        try:
            return inter_area / (self.area + other.area - inter_area)
        except ZeroDivisionError:
            return 0


class Box3d(ToNumpy):
    """A 3-D bounding box

    Attributes
    ----------
    center : Point
        Center :py:class:`Point` of the bounding box
    length : float
        The x dimention of the box
    width : float
        The y dimention of the box
    height : float
        The z dimention of the box
    heading : float
        The heading of the bounding box (in radians).  The heading is the angle
        required to rotate +x to the surface normal of the box front face. It is
        normalized to ``[-pi, pi)``.

    References
    ----------
    * Waymo Dataset Spec https://github.com/waymo-research/waymo-open-dataset/blob/master/waymo_open_dataset/label.proto
    """  # noqa: E501

    __UDT__ = Box3dType()

    def __init__(
        self,
        center: Point,
        length: float,
        width: float,
        height: float,
        heading: float,
    ):
        self.center = center
        self.length = float(length)
        self.width = float(width)
        self.height = float(height)
        self.heading = float(heading)

    def __repr__(self) -> str:
        return (
            f"Box3d(center={self.center}, l={self.length}, "
            + f"h={self.height}, w={self.width}, heading={self.heading})"
        )

    def __eq__(self, o: object) -> bool:
        return (
            isinstance(o, Box3d)
            and self.center == o.center
            and self.length == o.length
            and self.width == o.width
            and self.height == o.height
            and self.heading == o.heading
        )

    def to_numpy(self) -> np.ndarray:
        return None



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from torchvision import transforms as T

from rikai.contrib.torch.transforms.utils import uri_to_pil

__all__ = ["pre_processing", "post_processing"]

pre_processing = T.Compose(
    [
        uri_to_pil,
        T.Resize(256),
        T.CenterCrop(224),
        T.ToTensor(),
    ]
)


def post_processing(batch):
    results = []
    for predicts in batch:
        predict_result = {
            "boxes": [],
            "labels": [],
            "scores": [],
        }
        for box, label, score in zip(
            predicts["boxes"].tolist(),
            predicts["labels"].tolist(),
            predicts["scores"].tolist(),
        ):
            predict_result["boxes"].append(box)
            predict_result["labels"].append(label)
            predict_result["scores"].append(score)

        results.append(predict_result)
    return results



#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""This module makes :py:class:`numpy.ndarray` inter-operatable with rikai
from feature engineerings in `Spark <https://spark.apache.org/>`_ to be trained
in Tensorflow and Pytorch.

>>> # Feature Engineering in Spark
>>> from rikai import numpy as np
>>> df = spark.createDataFrame([Row(mask=np.array([1, 2, 3, 4]))])
>>> df.write.format("rikai").save("s3://path/to/features")

When use the rikai data in training, the serialized numpy data will be
automatically converted into the appropriate format, i.e.,
:py:class:`torch.Tensor` in Pytorch:

>>> from rikai.torch import DataLoader
>>> data_loader = DataLoader("s3://path/to/features")
>>> next(data_loader)
{"mask": tensor([1, 2, 3])}

"""

# Third Party
import numpy as np

# Rikai
from rikai.mixin import ToNumpy
from rikai.spark.types import NDArrayType

__all__ = ["wrap", "array", "empty"]


class ndarray(np.ndarray, ToNumpy):  # pylint: disable=invalid-name
    """This class extends numpy ndarray to be serialized in Spark."""

    __UDT__ = NDArrayType()

    def to_numpy(self) -> np.ndarray:
        """Convert to a pure numpy array for compatibility"""
        # TODO: optimize it for zero-copy
        return np.copy(self)


def wrap(data: np.ndarray) -> np.ndarray:
    """Wrap a numpy array to be able to work Spark and Parquet.

    Parameters
    ----------
    data : np.ndarray
        A raw numpy array

    Returns
    -------
    np.ndarray
        A Numpy array that is compatible with Spark User Defined Type.


    Example
    -------
    >>> import numpy as np
    >>> from rikai.numpy import wrap
    >>>
    >>> arr = np.array([1, 2, 3], dtype=np.int64)
    >>> df = spark.createDataFrame([Row(id=1, mask=wrap(arr))])
    >>> df.write.format("rikai").save("s3://foo/bar")
    """
    return data.view(ndarray)


def array(obj, *args, **kwargs) -> np.ndarray:
    """Create an numpy array using the same API as :py:func:`numpy.array`.

    See Also
    --------

    :py:func:`numpy.array`
    """
    return np.array(obj, *args, **kwargs).view(ndarray)


def empty(shape, dtype=float, order="C") -> np.ndarray:
    """Return an empty :py:class:`np.ndarray`.

    The returned array can be directly used in a Spark
    :py:class:`~pyspark.sql.DataFrame`.

    See Also
    --------

    :py:func:`numpy.empty`
    """
    return wrap(np.empty(shape, dtype=dtype, order=order))



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Geometry User Defined Types in Spark
"""

from __future__ import annotations

# Third-Party
from pyspark.sql import Row
from pyspark.sql.types import (
    DoubleType,
    StructField,
    StructType,
    UserDefinedType,
)

# Rikai
from rikai.logging import logger

__all__ = ["PointType", "Box3dType", "Box2dType"]


class Box2dType(UserDefinedType):
    """User defined type for the 2D bounding box."""

    @classmethod
    def sqlType(cls) -> StructType:
        return StructType(
            fields=[
                StructField("xmin", DoubleType(), False),
                StructField("ymin", DoubleType(), False),
                StructField("xmax", DoubleType(), False),
                StructField("ymax", DoubleType(), False),
            ]
        )

    @classmethod
    def module(cls) -> str:
        return "rikai.spark.types.geometry"

    @classmethod
    def scalaUDT(cls) -> str:
        return "org.apache.spark.sql.rikai.Box2dType"

    def serialize(self, obj: "rikai.types.geometry.Box2d"):
        """Serialize a Box2d into a PySpark Row"""
        return (
            obj.xmin,
            obj.ymin,
            obj.xmax,
            obj.ymax,
        )

    def deserialize(self, datum: Row) -> "rikai.types.geometry.Box2d":
        from rikai.types.geometry import Box2d

        if len(datum) < 4:
            logger.error(f"Deserialize box2d: not sufficient data: {datum}")

        return Box2d(*datum[:4])

    def simpleString(self) -> str:
        return "box2d"


class PointType(UserDefinedType):
    """Spark UDT for :py:class:`rikai.types.geometry.Point` class."""

    @classmethod
    def sqlType(cls) -> StructType:
        return StructType(
            fields=[
                StructField("x", DoubleType(), False),
                StructField("y", DoubleType(), False),
                StructField("z", DoubleType(), False),
            ]
        )

    @classmethod
    def module(cls) -> str:
        return "rikai.spark.types.geometry"

    @classmethod
    def scalaUDT(cls) -> str:
        return "org.apache.spark.sql.rikai.PointType"

    def serialize(self, obj: "Point"):
        """Serialize an :py:class:`PointType` into Spark Row"""
        return Row(x=obj.x, y=obj.y, z=obj.z)

    def deserialize(self, datum: Row) -> "Point":
        from rikai.types.geometry import Point

        if len(datum) < 3:
            logger.error(f"Deserialize Point: not sufficient data: {datum}")

        return Point(datum[0], datum[1], datum[2])

    def simpleString(self) -> str:
        return "PointType"


class Box3dType(UserDefinedType):
    """Spark UDT for :py:class:`~rikai.types.geometry.Box3d` class."""

    @classmethod
    def sqlType(cls) -> StructType:
        return StructType(
            fields=[
                StructField("center", PointType.sqlType(), False),
                StructField("length", DoubleType(), False),
                StructField("width", DoubleType(), False),
                StructField("height", DoubleType(), False),
                StructField("heading", DoubleType(), False),
            ]
        )

    @classmethod
    def module(cls) -> str:
        return "rikai.spark.types.geometry"

    @classmethod
    def scalaUDT(cls) -> str:
        return "org.apache.spark.sql.rikai.Box3dType"

    def serialize(self, obj: "Box3d"):
        """Serialize an Box3d into a Spark Row"""
        return Row(obj.center, obj.length, obj.width, obj.height, obj.heading)

    def deserialize(self, datum: Row) -> "Box3d":
        from rikai.types.geometry import Box3d

        if len(datum) < 5:
            logger.error(f"Deserialize Box3d: not sufficient data: {datum}")
        return Box3d(datum[0], datum[1], datum[2], datum[3], datum[4])

    def simpleString(self) -> str:
        return "Box3dType"



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""I/O related PySpark UDFs."""

# Third-party
from pyspark.sql.functions import udf
from pyspark.sql.types import StringType

# Rikai
from rikai.io import copy as _copy

__all__ = ["copy"]


@udf(returnType=StringType())
def copy(source: str, dest: str) -> str:
    """Copy a file from source to dest

    Parameters
    ----------
    source : str
        The source URI to copy from
    dest : str
        The destination uri or the destionation directory. If ``dest`` is
        a URI ends with a "/", it represents a directory.

    Return
    ------
    str
        Return the URI of destination.
    """
    return _copy(source, dest)



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Video related types and utils"""
from abc import ABC, abstractmethod

import cv2

from rikai.mixin import Displayable
from rikai.spark.types import SegmentType, VideoStreamType, YouTubeVideoType

__all__ = [
    "YouTubeVideo",
    "VideoStream",
    "VideoSampler",
    "SingleFrameSampler",
    "Segment",
]


class YouTubeVideo(Displayable):
    """
    Represents a YouTubeVideo, the basis of many open-source video data sets.
    This classes uses the ipython display library to integrate with jupyter
    notebook display and uses youtube-dl to download and create a
    VideoStream instance which represents a particular video stream file obj
    """

    __UDT__ = YouTubeVideoType()

    def __init__(self, vid: str):
        """
        Parameters
        ----------
        vid: str
            The youtube video id
        """
        self.vid = vid
        self.uri = "https://www.youtube.com/watch?v={0}".format(self.vid)
        self.embed_url = "http://www.youtube.com/embed/{0}".format(self.vid)

    def __repr__(self) -> str:
        return "YouTubeVideo({0})".format(self.vid)

    def display(self, width: int = 400, height: int = 300, **kwargs):
        """
        Visualization in jupyter notebook with custom options

        Parameters
        ----------
        width: int, default 400
            Width in pixels
        height: int, default 300
            Height in pixels
        kwargs: dict
            See :py:class:`IPython.display.YouTubeVideo` for other kwargs

        Returns
        -------
        v: IPython.display.YouTubeVideo
        """
        from IPython.display import YouTubeVideo

        return YouTubeVideo(self.vid, width=width, height=height, **kwargs)

    def _repr_html_(self):
        """default visualization in jupyter notebook cell"""
        return self.display()._repr_html_()

    def __eq__(self, other) -> bool:
        return isinstance(other, YouTubeVideo) and self.vid == other.vid

    def get_stream(
        self, ext: str = "mp4", quality: str = "worst"
    ) -> "VideoStream":
        """
        Get a reference to a particular stream

        Parameters
        ----------
        ext: str, default 'mp4'
            The preferred extension type to get. One of ['ogg', 'm4a', 'mp4',
            'flv', 'webm', '3gp']
            See: https://pythonhosted.org/Pafy/#Pafy.Stream.extension
        quality: str, default 'worst'
            Either 'worst' (lowest bitrate) or 'best' (highest bitrate)
            See: https://pythonhosted.org/Pafy/index.html#Pafy.Pafy.getbest

        Returns
        -------
        v: VideoStream
            VideoStream referencing an actual video resource
        """
        try:
            import pafy
        except ImportError as e:
            print(
                "Run `pip install rikai[youtube] to install pafy and "
                "youtube_dl to work with youtube videos."
            )
            raise e
        ext, quality = ext.strip().lower(), quality.strip().lower()
        if quality == "worst":
            stream = getworst(pafy.new(self.uri), preftype=ext)
        else:
            stream = pafy.new(self.uri).getbest(preftype=ext)
        return VideoStream(stream.url)


# Pafy hasn't had a new release with getworst yet
def getworst(v_pafy, preftype="any", ftypestrict=True, vidonly=False):
    """
    Return the highest resolution video available.

    Select from video-only streams if vidonly is True
    """
    streams = v_pafy.videostreams if vidonly else v_pafy.streams

    if not streams:
        return None

    def _sortkey(x, key3d=0, keyres=0, keyftype=0):
        """ sort function for max(). """
        key3d = "3D" not in x.resolution
        keyres = int(x.resolution.split("x")[0])
        keyftype = preftype == x.extension
        strict = (key3d, keyftype, keyres)
        nonstrict = (key3d, keyres, keyftype)
        return strict if ftypestrict else nonstrict

    r = min(streams, key=_sortkey)
    if ftypestrict and preftype != "any" and r.extension != preftype:
        return None
    return r


class VideoStream(Displayable):
    """Represents a particular video stream at a given uri"""

    __UDT__ = VideoStreamType()

    def __init__(self, uri: str):
        self.uri = uri

    def __repr__(self) -> str:
        return f"VideoStream(uri={self.uri})"

    def display(self, width: int = None, height: int = None, **kwargs):
        """
        Customize visualization in jupyter notebook

        Parameters
        ----------
        width: int, default None
            Width in pixels. Defaults to the original video width
        height: int, default None
            Height in pixels. Defaults to the original video height
        kwargs: dict
            See :py:class:`IPython.display.Video` doc for other kwargs

        Returns
        -------
        v: IPython.display.Video
        """
        from IPython.display import Video

        return Video(self.uri, width=width, height=height, **kwargs)

    def _repr_html_(self):
        """default visualizer for jupyter notebook"""
        return self.display()._repr_html_()

    def __eq__(self, other) -> bool:
        return isinstance(other, VideoStream) and self.uri == other.uri

    def __iter__(self):
        """Iterate through every frame in the video"""
        for frame in SingleFrameSampler(self):
            yield frame


class Segment:
    """A video segment bounded by frame numbers"""

    __UDT__ = SegmentType()

    def __init__(self, start_fno: int, end_fno: int):
        """

        Parameters
        ----------
        start_fno: int
            The starting frame number (0-indexed)
        end_fno: int
            The ending frame number. If <0 then it means end of the video

        Notes
        -----
        `fno` terminology is chosen to be consistent with the opencv library.
        """
        if start_fno < 0:
            raise ValueError("Cannot start with negative frame number")
        if end_fno > 0 and end_fno < start_fno:
            raise ValueError(
                "Ending frame must be negative or larger than starting frame"
            )
        self.start_fno = start_fno
        self.end_fno = end_fno

    def __repr__(self) -> str:
        return f"Segment(start_fno={self.start_fno}, end_fno={self.end_fno})"

    def __eq__(self, other) -> bool:
        return isinstance(other, Segment) and (
            self.start_fno,
            self.end_fno,
        ) == (
            other.start_fno,
            other.end_fno,
        )


class VideoSampler(ABC):
    """
    Subclasses will implement different ways to retrieve samples from a given
    VideoStream.
    """

    def __init__(self, stream: VideoStream):
        self.stream = stream

    @abstractmethod
    def __iter__(self):
        pass


class SingleFrameSampler(VideoSampler):
    """
    A simple sampler that just returns one out of every `sample_rate` frames
    """

    def __init__(
        self,
        stream: VideoStream,
        sample_rate: int = 1,
        start_frame: int = 0,
        max_samples: int = -1,
    ):
        """
        Parameters
        ----------
        sample_rate: int
            The sampling rate in number of frames
        start_frame: int
            Start from a specific frame (0-based indexing)
        max_samples: int
            Yield at most this many frames (-1 means no max)
        """
        super().__init__(stream)
        self.sample_rate = sample_rate
        self.start_frame = start_frame
        self.max_samples = max_samples

    def __iter__(self):
        # TODO use seek for sparse sampling and maybe multithreaded
        cap = cv2.VideoCapture(self.stream.uri)
        if self.start_frame > 0:
            cap.set(cv2.CAP_PROP_POS_FRAMES, self.start_frame)
        fno = 0
        success = cap.grab()  # get the next frame
        tot_samples = 0
        while success:
            if fno % self.sample_rate == 0:
                _, img = cap.retrieve()
                tot_samples += 1
                yield img
            if self.max_samples > -1 and tot_samples >= self.max_samples:
                break
            success = cap.grab()



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pyspark.sql import SparkSession

from rikai.spark.sql.callback_service import init_cb_service

__all__ = ["init"]


def init(spark: SparkSession):
    """Initialize SQL-ML services.

    .. code-block:: python

        # To start use Rikai SQL-ML in Spark
        from rikai.spark.sql import init

        spark = (
            SparkSession
            ...
            .getOrCreate)
        init(spark)

    """
    init_cb_service(spark)



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Vision related Spark UDFs.
"""

# Standard library
import os
from typing import Union

# Third Party
import numpy as np
from pyspark.sql.functions import udf
from pyspark.sql.types import ArrayType

# Rikai
from rikai.io import copy as _copy
from rikai.logging import logger
from rikai.numpy import ndarray
from rikai.spark.types.vision import ImageType
from rikai.types.video import (
    Segment,
    SingleFrameSampler,
    VideoStream,
    YouTubeVideo,
)
from rikai.types.vision import Image

__all__ = [
    "image",
    "image_copy",
    "numpy_to_image",
    "video_to_images",
    "spectrogram_image",
]


@udf(returnType=ImageType())
def image(uri: str) -> Image:
    """Build an :py:class:`Image` from a URI."""
    return Image(uri)


@udf(returnType=ImageType())
def image_copy(img: Image, uri: str) -> Image:
    """Copy the image to a new destination, specified by the URI.

    Parameters
    ----------
    img : Image
        An image object
    uri : str
        The base directory to copy the image to.

    Return
    ------
    Image
        Return a new image pointed to the new URI
    """
    logger.info("Copying image src=%s dest=%s", img.uri, uri)
    return Image(_copy(img.uri, uri))


@udf(returnType=ImageType())
def numpy_to_image(array: ndarray, uri: str) -> Image:
    """Convert a numpy array to image, and upload to external storage.

    Parameters
    ----------
    array : :py:class:`numpy.ndarray`
        Image data.
    uri : str
        The base directory to copy the image to.

    Return
    ------
    Image
        Return a new image pointed to the new URI.

    Example
    -------

    >>> spark.createDataFrame(..).registerTempTable("df")
    >>>
    >>> spark.sql(\"\"\"SELECT numpy_to_image(
    ...        resize(grayscale(image)),
    ...        lit('s3://asset')
    ...    ) AS new_image FROM df\"\"\")

    See Also
    --------
    :py:meth:`rikai.types.vision.Image.from_array`
    """
    return Image.from_array(array, uri)


@udf(returnType=ArrayType(ImageType()))
def video_to_images(
    video: Union[VideoStream, YouTubeVideo],
    output_uri: str,
    segment: Segment = Segment(0, -1),
    sample_rate: int = 1,
    max_samples: int = 15000,
    quality: str = "worst",
) -> list:
    """Extract video frames into a list of images.

    Parameters
    ----------
    video : Video
        An video object, either YouTubeVideo or VideoStream.
    output_uri: str
        Frames will be written as <output_uri>/<fno>.jpg
    segment: Segment, default Segment(0, -1)
        A Segment object, localizing video in time to (start_fno, end_fno)
    sample_rate : int, default 1
        Keep 1 out of every sample_rate frames.
    max_samples : int, default 15000
        Return at most this many frames (-1 means no max)
    quality: str, default 'worst'
        Either 'worst' (lowest bitrate) or 'best' (highest bitrate)
        See: https://pythonhosted.org/Pafy/index.html#Pafy.Pafy.getbest

    Return
    ------
    List
        Return a list of images from video indexed by frame number.
    """
    assert isinstance(
        video, (YouTubeVideo, VideoStream)
    ), "Input type must be YouTubeVideo or VideoStream"
    assert isinstance(segment, Segment), "Second input type must be Segment"

    start_frame = segment.start_fno
    if segment.end_fno > 0:
        max_samples = min((segment.end_fno - start_frame), max_samples)

    if isinstance(video, YouTubeVideo):
        video_iterator = SingleFrameSampler(
            video.get_stream(quality=quality),
            sample_rate,
            start_frame,
            max_samples,
        )
    else:
        video_iterator = SingleFrameSampler(
            video, sample_rate, start_frame, max_samples
        )

    return [
        Image.from_array(
            img,
            os.path.join(
                output_uri, "{}.jpg".format((start_frame + idx) * sample_rate)
            ),
        )
        for idx, img in enumerate(video_iterator)
    ]


@udf(returnType=ImageType())
def spectrogram_image(
    video: Union[VideoStream, YouTubeVideo],
    output_uri: str,
    segment: Segment = Segment(0, -1),
    size: int = 224,
    max_samples: int = 15000,
) -> Image:
    """Applies ffmpeg filter to generate spectrogram image.

    Parameters
    ----------
    video : VideoStream or YouTubeVideo
        A video object whose audio track will be converted to spectrogram
    output_uri: str
        The uri to which the spectrogram image will be written to
    segment: Segment
            A Segment object, localizing video in time to (start_fno, end_fno)
    max_samples : Int
            Yield at most this many frames (-1 means no max)
    size : Int
        Sets resolution of frequency, time spectrogram image.

    Return
    ------
    Image
        Return an Image of the audio spectrogram.
    """
    try:
        import ffmpeg
    except ImportError:
        raise ValueError(
            "Couldn't import ffmpeg. Please make sure to "
            "`pip install ffmpeg-python` explicitly or install "
            "the correct extras like `pip install rikai[all]`"
        )
    assert isinstance(
        video, (YouTubeVideo, VideoStream)
    ), "Input type must be YouTubeVideo or VideoStream"
    assert isinstance(segment, Segment), "Second input type must be Segment"

    start_frame = segment.start_fno
    if segment.end_fno > 0:
        max_samples = min((segment.end_fno - start_frame), max_samples)
    video_uri = (
        video.get_stream().uri
        if isinstance(video, YouTubeVideo)
        else video.uri
    )
    output, _ = (
        ffmpeg.input(video_uri)
        .filter("showspectrumpic", "{}x{}".format(size, size), legend=0)
        .output(
            "pipe:",
            format="rawvideo",
            pix_fmt="rgb24",
            start_number=start_frame,
            vframes=max_samples,
        )
        .run(capture_stdout=True)
    )
    return Image.from_array(
        np.frombuffer(output, np.uint8).reshape([size, size, 3]), output_uri
    )



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# Standard
from typing import Any, Callable, Optional, Union

# Third Party
import pandas as pd
from torch.utils.data import Dataset

# Rikai
from rikai.torch.transforms import convert_tensor

__all__ = ["PandasDataset"]


class PandasDataset(Dataset):
    """a Map-style Pytorch dataset from a :py:class:`pandas.DataFrame` or a
    :py:class:`pandas.Series`.

    Note
    ----

    This class is used in Rikai's SQL-ML Spark implementation, which utilizes
    pandas UDF to run inference.
    """

    def __init__(
        self,
        data: Union[pd.DataFrame, pd.Series],
        transform: Optional[Callable] = None,
    ) -> None:
        assert isinstance(data, (pd.DataFrame, pd.Series))
        self.data = data
        self.transform = transform

    def __len__(self) -> int:
        return self.data.shape[0]

    def __getitem__(self, index: int) -> Any:
        row = self.data.iloc[index]
        row = convert_tensor(row)
        if self.transform:
            row = self.transform(row)
        return row



#  Copyright 2020 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Mixins
"""
from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path
from typing import BinaryIO, Union

# Third Party
import numpy as np

# Rikai
from rikai.internal.uri_utils import uri_equal
from rikai.io import open_uri

__all__ = ["ToNumpy", "Asset", "Displayable"]


class ToNumpy(ABC):
    """ToNumpy Mixin."""

    @abstractmethod
    def to_numpy(self) -> np.ndarray:
        """Returns the content as a numpy ndarray."""


class ToPIL(ABC):
    """ToPIL Mixin."""

    @abstractmethod
    def to_pil(self) -> "PIL.Image.Image":
        pass


class Displayable(ABC):
    """Mixin for notebook viz"""

    @abstractmethod
    def display(self, **kwargs) -> "IPython.display.DisplayObject":
        """Return an IPython.display.DisplayObject"""


class Asset(ABC):
    """cloud asset Mixin.

    Rikai uses asset to store certain blob on the cloud storage, to facilitate
    the functionality like fast query, example inspections, and etc.

    An asset is also a cell in a DataFrame for analytics. It offers both fast
    query on columnar format and easy tooling to access the actual data.
    """

    def __init__(self, uri: Union[str, Path]) -> None:
        self.uri = str(uri)

    def __eq__(self, o: object) -> bool:
        return isinstance(o, Asset) and uri_equal(self.uri, o.uri)

    def open(self, mode="rb") -> BinaryIO:
        """Open the asset and returned as random-accessible file object."""
        return open_uri(self.uri, mode=mode)



import pathlib
import re

from setuptools import find_packages, setup

about = {}
with open(pathlib.Path("rikai") / "__version__.py", "r") as fh:
    exec(fh.read(), about)

with open(
    pathlib.Path(__file__).absolute().parent.parent / "README.md", "r"
) as fh:
    long_description = fh.read()

# extras
test = ["pytest"]
torch = ["torch>=1.4.0", "torchvision"]
jupyter = ["matplotlib", "jupyterlab"]
aws = ["boto"]
gcp = ["gcsfs"]
docs = ["sphinx"]
youtube = ["pafy", "youtube_dl", "ffmpeg-python"]
all = test + torch + jupyter + aws + gcp + docs + youtube


setup(
    name="rikai",
    version=about["version"],
    license="Apache License, Version 2.0",
    author="Rikai authors",
    author_email="rikai-dev@eto.ai",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/eto-ai/rikai",
    packages=find_packages(),
    include_package_data=True,
    python_requires=">=3.7",
    install_requires=[
        "antlr4-python3-runtime",
        "ipython",
        "jsonschema",
        "numpy",
        "opencv-python",
        "pandas",
        "Pillow",
        "pyarrow>=2.0",
        "pyspark>=3.1,<3.2",
        "pyyaml",
        "requests",
        "semver",
    ],
    extras_require={
        "test": test,
        "pytorch": torch,
        "jupyter": jupyter,
        "aws": aws,
        "docs": docs,
        "youtube": youtube,
        "all": all,
    },
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries",
    ],
)



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from py4j.java_collections import JavaMap
from py4j.java_gateway import CallbackServerParameters
from pyspark.sql import SparkSession

from rikai.logging import logger

__all__ = ["init_cb_service"]


def init_cb_service(spark: SparkSession):
    jvm = spark.sparkContext._gateway
    params = CallbackServerParameters(
        daemonize=True,
        daemonize_connections=True,
        # Re-use the auth-token from the main java/spark process
        auth_token=jvm.gateway_parameters.auth_token,
    )
    jvm.start_callback_server(callback_server_parameters=params)
    logger.info("Spark callback server started")

    cb = CallbackService(spark)
    cb.register()


class CallbackService:
    """:py:class:`CallbackService` allows SparkSessions' JVM to run
    arbitrary code in SparkSession's python interpreter.

    Notes
    -----
    Internal use only
    """

    def __init__(self, spark: SparkSession):
        self.spark = spark
        self.registry_map = {}

    def __repr__(self):
        return "PythonCbService"

    def resolve(
        self, registry_class: str, uri: str, name: str, options: JavaMap
    ):
        """Resolve a ML model.

        Parameters
        ----------
        registry_class : str
            The full class name for the registry.
        uri : str
            The model URI.
        name : str
            Mode name
        options : dict[str, str]
            Options passed to the model.
        """
        if registry_class not in self.registry_map:
            from rikai.internal.reflection import find_class
            from rikai.spark.sql.codegen.base import Registry

            cls = find_class(registry_class)
            if not issubclass(cls, Registry):
                raise ValueError(
                    f"Class '{registry_class}' is not a Registry'"
                )
            self.registry_map[registry_class] = cls(self.spark)

        registry = self.registry_map[registry_class]
        # Convert JavaMap to dict
        options = {key: options[key] for key in options.keys()}
        return registry.resolve(uri, name, options)

    def register(self):
        """Register this :py:class:`CallbackService` to SparkSession's JVM."""
        jvm = self.spark.sparkContext._jvm
        jvm.ai.eto.rikai.sql.spark.Python.register(self)
        logger.info(
            "Rikai Python CallbackService is registered to SparkSession"
        )

    def toString(self):
        """For Java compatibility"""
        return repr(self)

    class Java:
        implements = ["ai.eto.rikai.sql.spark.Python"]



#  Copyright 2020 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from antlr4 import CommonTokenStream, InputStream
from pyspark.sql.types import (
    ArrayType,
    BinaryType,
    BooleanType,
    ByteType,
    DataType,
    DoubleType,
    FloatType,
    IntegerType,
    LongType,
    ShortType,
    StringType,
    StructField,
    StructType,
)

from rikai.spark.sql.generated.RikaiModelSchemaLexer import (
    RikaiModelSchemaLexer,
)
from rikai.spark.sql.generated.RikaiModelSchemaParser import (
    RikaiModelSchemaParser,
)
from rikai.spark.sql.generated.RikaiModelSchemaVisitor import (
    RikaiModelSchemaVisitor,
)

_SPARK_TYPE_MAPPING = {
    "bool": BooleanType(),
    "boolean": BooleanType(),
    "byte": ByteType(),
    "tinyint": ByteType(),
    "short": ShortType(),
    "smallint": ShortType(),
    "int": IntegerType(),
    "long": LongType(),
    "bigint": LongType(),
    "float": FloatType(),
    "double": DoubleType(),
    "str": StringType(),
    "string": StringType(),
    "binary": BinaryType(),
}


class SchemaError(Exception):
    def __init__(self, message: str):
        self.message = message


class SchemaBuilder(RikaiModelSchemaVisitor):
    def visitStructType(
        self, ctx: RikaiModelSchemaParser.StructTypeContext
    ) -> StructType:
        return StructType(
            [self.visitStructField(field) for field in ctx.field()]
        )

    def visitStructField(
        self, ctx: RikaiModelSchemaParser.StructFieldContext
    ) -> StructField:
        name = self.visit(ctx.identifier())
        dataType = self.visit(ctx.fieldType())
        return StructField(name, dataType)

    def visitArrayType(
        self, ctx: RikaiModelSchemaParser.ArrayTypeContext
    ) -> ArrayType:
        return ArrayType(self.visit(ctx.fieldType()))

    def visitUnquotedIdentifier(
        self, ctx: RikaiModelSchemaParser.UnquotedIdentifierContext
    ) -> str:
        identifer = ctx.IDENTIFIER().getText()
        if identifer[0].isnumeric():
            raise SchemaError(
                f'Identifier can not start with a digit: "{identifer}"'
            )
        return identifer

    def visitPlainFieldType(
        self, ctx: RikaiModelSchemaParser.PlainFieldTypeContext
    ) -> DataType:
        name = self.visit(ctx.identifier())
        try:
            return _SPARK_TYPE_MAPPING[name]
        except KeyError as e:
            # TODO: Support customized UDT
            raise SchemaError(f'Can not recognize type: "{name}"') from e


def parse_schema(schema_str: str) -> DataType:
    input_stream = InputStream(schema_str)
    lexer = RikaiModelSchemaLexer(input_stream)
    stream = CommonTokenStream(lexer)
    parser = RikaiModelSchemaParser(stream)

    visitor = SchemaBuilder()
    schema = visitor.visit(parser.schema())
    # TODO(GH#112) we should add error listener to Antlr Parser.
    if schema is None:
        raise SchemaError(f"Invalid schema: '{schema_str}'")
    return schema



#  Copyright 2020 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Pytorch Dataset and DataLoader"""
import os
import threading
import uuid
import warnings
from concurrent.futures import as_completed, ThreadPoolExecutor
from pathlib import Path
from queue import Empty, Queue
from typing import Callable, Dict, Generator, List, Optional, Union

# Third Party
import semver
import torch
from torch.utils.data import IterableDataset

# Rikai
import rikai.parquet
from rikai.torch.transforms import convert_tensor, RikaiToTensor

__all__ = ["DataLoader", "Dataset"]


class Dataset(IterableDataset):
    """Rikai Pytorch Dataset.

    A :py:class:`torch.utils.data.IterableDataset` that reads
    Rikai data format. This :py:class:`Dataset` works with
    `multi-process data loading`_ using :py:class:`torch.utils.data.DataLoader`.

    Parameters
    ----------
    data_ref : str, Path, pyspark.sql.DataFrame
        URI to the data files or the dataframe
    columns : list of str, optional
        An optional list of column to load from parquet files.
    transform: Callable, default instance of RikaiToTensor
        Apply row level transformation before yielding each sample

    Note
    ----

    Up to ``pytorch==1.7``, :py:class:`~torch.utils.data.IterableDataset`
    does not work with :py:class:`torch.utils.data.Sampler` with
    :py:class:`torch.utils.data.DataLoader`.

    Use :py:class:`torch.utils.data.BufferedShuffleDataset` (torch>=1.8)
    with the Rikai dataset for randomness.

    Example
    -------

    >>> from rikai.torch.data import Dataset
    >>> from torch.utils.data import DataLoader
    >>>
    >>> dataset = Dataset("dataset", columns=["image", "label"])
    >>> # dataset = BufferedShuffleDataset(dataset)
    >>> loader = DataLoader(dataset, num_workers=8)

    .. _multi-process data loading: https://pytorch.org/docs/master/data.html#single-and-multi-process-data-loading
    """  # noqa: E501

    def __init__(
        self,
        data_ref: Union[str, Path, "pyspark.sql.DataFrame"],
        columns: List[str] = None,
        transform: Callable = RikaiToTensor(),
    ):
        super().__init__()
        self.data_ref = data_ref
        self.columns = columns
        self._transform = transform

    def __repr__(self) -> str:
        return f"Dataset(torch, {self.data_ref}, columns={self.columns})"

    def __iter__(self):
        rank = 0
        world_size = 1

        worker_info = torch.utils.data.get_worker_info()
        if worker_info is not None:
            rank = worker_info.id
            world_size = worker_info.num_workers

        uri = _maybe_cache_df(self.data_ref)
        for row in rikai.parquet.Dataset(
            uri,
            columns=self.columns,
            world_size=world_size,
            rank=rank,
        ):
            yield self._transform(row)


class DataLoader:
    """Rikai Dataset Loader in Pytorch.

    Parameters
    ----------
    data_ref : str
        The URI of a Rikai Dataset
    columns : list of str, optional
        An optional list of column to load from parquet files.
    batch_size : int
        Batch size.
    shuffle : bool
        Whether to shuffle the dataset or not. Default is False.
    num_workers : int
        The number of workers to download asset in parallel.
    seed : int, optional
        Provide random seed for shuffling process
    world_size : int
        The total number of distributed workers. Default is ``1``.
    rank : int
        The rank of this worker among distributed workers. Default is ``0``.

    **Distributed Training**

    :py:class:`DataLoader` can work with distributed training framework,
    such as `Horovod <https://horovod.readthedocs.io/en/stable/pytorch.html>`_.

    .. code-block:: python

        import torch
        import horovod.torch as hvd
        from rikai.torch.data import DataLoader

        # Initialize Horovod
        hvd.init()

        # Partition the dataset using Horovod primitives.
        train_loader = DataLoader(
            "s3://dataset/train",
            batch_size=16,
            world_size=hvd.size(),  # Horovod cluster size
            rank=hvd.rank())  # Local rank

        # Set ups on https://horovod.readthedocs.io/en/stable/pytorch.html

        for epoch in range(100):
            for batch_idx, (data, target) in enumerate(train_loader):
                ...

    .. warning::

        With Pytorch 1.8+, users should use the official :py:class:`torch.utils.data.DataLoader`
        with :py:class:`torch.utils.data.BufferedShuffleDataset` instead.

        This class will be deprecated later.

    References
    ----------
    .. `Horovod with Pytorch <https://horovod.readthedocs.io/en/stable/pytorch.html>`_

    """  # noqa

    def __init__(
        self,
        data_ref: Union[str, Path, "pyspark.sql.DataFrame"],
        columns: List[str] = None,
        batch_size: int = 1,
        shuffle: bool = False,
        num_workers: int = 16,
        collate_fn: Callable = None,
        seed: Optional[int] = None,
        world_size: int = 1,
        rank: int = 0,
    ):  # pylint: disable=too-many-arguments
        assert isinstance(data_ref, (str, Path)) or (
            rank == 0 and world_size == 1
        ), "Only str/Path references are supported in distributed mode"
        data_ref = _maybe_cache_df(data_ref)
        self.dataset = rikai.parquet.Dataset(
            data_ref,
            columns=columns,
            shuffle=shuffle,
            seed=seed,
            world_size=world_size,
            rank=rank,
        )
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.collate_fn = collate_fn if collate_fn else lambda x: x

        torch_version = semver.VersionInfo.parse(torch.__version__)
        if torch_version.major >= 1 and torch_version.minor >= 8:
            warnings.warn(
                "rikai.torch.data.DataLoader should be replaced with "
                "'torch.utils.data.BufferedShuffleDataset' and "
                "'torch.utils.data.DataLoader' "
                "in Pytorch 1.8+",
                DeprecationWarning,
            )

    def _prefetch(
        self, out_queue: Queue, done: threading.Event, stop: threading.Event
    ) -> None:
        """Prefetch dataset in a separate threads.

        Prefetch and convert the dataset to tensor, including download assets
        in the background.

        Parameters
        ----------
        loader : DataLoader
            The DataLoader object
        out_queue : Queue
            A thread-safe queue to collect converted rows.
        done : threading.Event
            Signal other threads that the prefetch is completed
        stop : threading.Event
            Signal from the calling thread to stop fetching data
        """
        try:

            def parallel_prefetch(executor, batch, q):
                futures = [executor.submit(convert_tensor, e) for e in batch]
                for fut in as_completed(futures):
                    q.put(fut.result())

            prefetch_batch = 4 * self.num_workers
            with ThreadPoolExecutor(self.num_workers) as executor:
                batch = []
                for example in self.dataset:
                    batch.append(example)
                    if len(batch) > prefetch_batch:
                        parallel_prefetch(executor, batch, out_queue)
                        batch = []
                    if stop.is_set():
                        break

                if batch:
                    parallel_prefetch(executor, batch, out_queue)
            out_queue.join()
        finally:
            # Signal the main thread that all datasat has been produced.
            done.set()

    def __iter__(self) -> Generator[Dict, None, None]:
        """Use DataLoader as a generator of example."""
        q = Queue(maxsize=self.num_workers * self.batch_size * 4)
        done = threading.Event()
        stop = threading.Event()

        prefetch_thd = threading.Thread(
            target=DataLoader._prefetch,
            args=(self, q, done, stop),
            daemon=True,
        )
        prefetch_thd.start()

        try:
            batch = []
            while not (q.empty() and done.is_set()):
                try:
                    batch.append(q.get(timeout=5))
                    q.task_done()
                    if len(batch) >= self.batch_size:
                        yield self.collate_fn(batch)
                        batch = []
                except Empty:
                    continue
            if batch:
                yield self.collate_fn(batch)
        finally:
            stop.set()  # tell the prefetch thread to stop
        prefetch_thd.join()
        q.join()


def _maybe_cache_df(
    data_ref: Union[str, Path, "pyspark.sql.DataFrame"]
) -> str:
    """
    If the given dataset_ref is a str/Path, then just return a str ref.
    If it's a pyspark DataFrame then cache the DataFrame as parquet and
    return the cache uri

    Parameters
    ----------
    data_ref: str, Path, or pyspark DataFrame
        Either a uri to parquet or a DataFrame that will be cached as parquet

    Returns
    -------
    uri: str
        Either the input parquet uri or the cached uri for a DataFrame
    """
    if isinstance(data_ref, (str, Path)):
        return str(data_ref)
    else:
        try:
            from pyspark.sql import DataFrame
        except ImportError:
            raise ImportError(
                "Cannot create rikai pytorch dataset from "
                "Spark DataFrame without pyspark installed."
            )
        from rikai.spark.utils import CONF_PARQUET_BLOCK_SIZE, df_to_rikai

        if isinstance(data_ref, DataFrame):
            kwargs = {}
            # TODO also try to get it from rikai.options once #134 is done
            block_size = data_ref.rdd.ctx.getConf().get(
                CONF_PARQUET_BLOCK_SIZE, None
            )
            if block_size is not None:
                kwargs["parquet_row_group_size_bytes"] = block_size
            cache_uri = _get_cache_uri(data_ref)
            df_to_rikai(data_ref, cache_uri, **kwargs)
        else:
            raise TypeError(
                (
                    "dataset_ref must be a str, Path, or DataFrame and "
                    "not a {}."
                ).format(type(data_ref))
            )
        return cache_uri


def _get_cache_uri(df: "pyspark.sql.DataFrame") -> str:
    """
    TODO create a deterministic unique uri from the df for sharing

    Parameters
    ----------
    df: DataFrame
        The cache uri will be generated for the given DataFrame

    Returns
    -------
    cache_uri: str
        The uri to write the DataFrame to
    """
    # TODO also try to get it from rikai.options once #134 is done
    from rikai.spark.conf import CONF_SPARK_RIKAI_CACHEURI

    cache_root_uri = df.rdd.ctx.getConf().get(CONF_SPARK_RIKAI_CACHEURI)
    return os.path.join(cache_root_uri, str(uuid.uuid4()))



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Helper functions for using :py:func:`assert` with pytests
"""

from typing import Iterable


def assert_count_equal(first: Iterable, second: Iterable, msg=None):
    """Assert ``first`` has the same elements as ``second``, regardless of
    the order.

    See Also
    --------
    :py:meth:`unittest.TestCase.assertCountEqual`
    """
    from unittest import TestCase

    TestCase().assertCountEqual(first, second, msg=msg)
